import sys
import copy
from pysat.solvers import Glucose3
from itertools import product

sudoku = []

# Copy one sudoku from https://github.com/dimitri/sudoku/blob/master/sudoku.txt into a file

f = open(sys.argv[1],'r') 
print("Loading the following sudoku: \n")

for fline in f:
    line = [int(d) for d in fline[:-1]]
    sudoku.append(line)
    print(line)

# sudoku now contains the input sudoku


def getConstraints(ps): # returns the set of lines, rows, and blocks
    constraints = copy.copy(ps) # All the lines 
    for j in range(9): # Extract column j
        row = []
        for i in range(9):
            row.append(ps[i][j])
        constraints.append(row)
    for i in range(3): 
        for j in range(3): # Extract the block (i,j)
            block = []
            for k in range(3):
                for l in range(3):
                    block.append(ps[3*i + k][3*j + l])
            constraints.append(block)
    return constraints

print("\nThe constraints are:\n")
for constr in getConstraints(sudoku):
    print(constr)


def firstUndefined(ps):
    'Returns the first position (i,j) containing a zero, or (-1,-1) if it is already full'
    for i in range(9):
        for j in range(9):
            if ps[i][j] == 0:
                return (i,j)
    return (-1,-1)


def noProblemoSoFar(ps):
    '''Returns False if the partial sudoku ps contains a repeated number (not
    including 0) in a row, column or block, and True otherwise'''
    for constraint in getConstraints(ps):
        for n in range(1,10):
            seen = False
            for v in constraint:
                if v == n:
                    if not seen:
                        seen = True
                    else:
                        return False
    return True



def solveBruteforce(ps):
    '''Tries to solve the sudoku by going over all possible values for all the
    missing squares. We go over all possible assignments of the missing squares
    and we check wether the obtained full sudoku satisfies all the constraints'''

    def recurseSolve(ps): # return (True, solution) if there is a solution, (False, False) otherwise
        i,j = firstUndefined(ps)
        if (i,j) == (-1,-1):
            if noProblemoSoFar(ps):
                return (True, ps)
            else:
                return (False, False)
        for value in range(1,10):
            nextps = copy.deepcopy(ps)
            nextps[i][j] = value
            status, sol = recurseSolve(nextps)
            if status:
                return (True, sol)
        return (False, False)
            

    status, sol = recurseSolve(ps)
    if status:
        print("\nHere is a solution:\n")
        for l in sol:
            print(l)
    else:
        print("No solution :(\n")



def solveUsingNondeterminism(ps):
    'Tries to solve the sudoku using the nondet package of Charles'
    pass



def solveBacktracking(ps):
    '''Tries to solve the sudoku using Backtracking. See the Wikipedia page on
    Backtracking (https://en.wikipedia.org/wiki/Backtracking#Pseudocode). The
    difference with bruteforce is that, instead of blindly going over all
    possible assignments of the missing values, we assign them one by one and
    we do not continue the exploration when the partial sudoku already violates
    a constraint'''
    def recurseSolve(ps): # return (True, solution) if there is a solution, (False, False) otherwise
        if not noProblemoSoFar(ps):
            return (False,False)
        i,j = firstUndefined(ps)
        if (i,j) == (-1,-1):
            return (True, ps)
        for value in range(1,10):
            nextps = copy.deepcopy(ps)
            nextps[i][j] = value
            status, sol = recurseSolve(nextps)
            if status:
                return (True, sol)
        return (False, False)
            

    status, sol = recurseSolve(ps)
    if status:
        print("\nHere is a solution:\n")
        for l in sol:
            print(l)
    else:
        print("No solution :(\n")



def solveUsingSATsolver(ps):
    '''Tries to solve the sudoku using the SAT solver PySAT
    (see https://pysathq.github.io/usage/ for how to use it)'''

    d1 = {}   # d1[(i,j,v)] is the integer used for variable (i,j,v) to give to the SAT solver
    # global current 
    current = 1

    def myvartoint(i,j,v):
        nonlocal current
        if (i,j,v) in d1:
            return d1[(i,j,v)]
        d1[(i,j,v)]  = current
        current += 1
        return current-1

    g = Glucose3()

    for i in range(1,10):
        for j in range(1,10):
            # clauses expressing that each cell receives *at least* one value
            g.add_clause([myvartoint(i,j,v) for v in range(1,10)])
            # clauses expressing that each cell receives *at most* one value
            for v in range(1,10):
                for vp in range(1,10):
                    if v != vp:
                        g.add_clause([-myvartoint(i,j,v), -myvartoint(i,j,vp)])

    # clauses indicating what are the values of the cells that are already filled
    for i in range(9):
        for j in range(9):
            if ps[i][j] != 0:
                g.add_clause([ myvartoint(i+1,j+1,ps[i][j])])

    # clauses indicating that for every row i and every value v the value v should appear 
    for i in range(1,10):
        for v in range(1,10):
            g.add_clause([myvartoint(i,j,v) for j in range(1,10)])

    # clauses indicating that for every columns i and every value v the value v should appear
    for j in range(1,10):
        for v in range(1,10):
            g.add_clause([myvartoint(i,j,v) for i in range(1,10)])

    # clauses indicating that for block and every value v the value v should appear
    for b1 in range(3):
        for b2 in range(3):
            for v in range(1,10):
                g.add_clause([  myvartoint(3*b1 + t1 + 1, 3*b2 + t2 +1, v) for (t1,t2) in product(range(3), repeat=2)])

    print(g.solve())
    if(g.solve()):
        print("A solution:\n")
        # We recover the solution and print the solved soduku
        model = g.get_model()
        for i in range(1,10):
            for j in range(1,10):
                for v in range(1,10):
                    if model[myvartoint(i,j,v) -1] > 0:
                        ps[i-1][j-1] = v
        for line in ps:
            print(line)
        
    else:
        print("No solution")



        




    # g.add_clause([-1, 2])
    # g.add_clause([-2, 3])
    # print(g.solve())
    # print(g.get_model())
    pass


# solveBruteforce(sudoku)
''' We observe that the bruteforce approach doesn't work well:
    If there are 40 zeros in the original sudoku and we start guessing
    the wrong value for the first zero, it will have to explore at least 9^39
    partial sudokus, which is way too much'''

# solveBacktracking(sudoku)
''' This on the other hand seems to work well. On the last example of
    the sudoku file https://github.com/dimitri/sudoku/blob/master/sudoku.txt,
    it still takes about 10s to solve.
'''

solveUsingSATsolver(sudoku)
''' This is crazy fast!
'''










